@@ -1,9 +1,9 @@
 /*
- *  $Id: buttons.c,v 1.76 2007/02/27 21:05:27 tom Exp $
+ *  $Id: buttons.c,v 1.81 2010/04/28 20:57:29 tom Exp $
  *
  * buttons.c -- draw buttons, e.g., OK/Cancel
  *
- * Copyright 2000-2006,2007 Thomas E. Dickey
+ * Copyright 2000-2007,2010 Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -58,19 +58,19 @@ string_to_char(const char **stringp)
     int result;
 #ifdef USE_WIDE_CURSES
     const char *string = *stringp;
-    int have = strlen(string);
-    int check;
-    int len;
+    size_t have = strlen(string);
+    size_t check;
+    size_t len;
     wchar_t cmp2[2];
     mbstate_t state;
 
     memset(&state, 0, sizeof(state));
     len = mbrlen(string, have, &state);
-    if (len > 0 && len <= have) {
+    if ((int) len > 0 && len <= have) {
 	memset(&state, 0, sizeof(state));
 	memset(cmp2, 0, sizeof(cmp2));
 	check = mbrtowc(cmp2, string, len, &state);
-	if (check <= 0)
+	if ((int) check <= 0)
 	    cmp2[0] = 0;
 	*stringp += len;
     } else {
@@ -142,7 +142,7 @@ print_button(WINDOW *win, char *label, int y, int x, int selected)
 	     ? button_active_attr
 	     : button_inactive_attr);
     (void) waddstr(win, ">");
-    (void) wmove(win, y, x + strspn(label, " ") + 1);
+    (void) wmove(win, y, x + ((int) strspn(label, " ")) + 1);
 }
 
 /*
@@ -204,6 +204,8 @@ dlg_button_x_step(const char **labels, int limit, int *gap, int *margin, int *st
     int unused;
     int used;
 
+    if (count == 0)
+	return 0;
     dlg_button_sizes(labels, FALSE, &longest, &length);
     used = (length + (count * 2));
     unused = limit - used;
@@ -228,13 +230,15 @@ dlg_button_layout(const char **labels, int *limit)
     int width = 1;
     int gap, margin, step;
 
-    while (!dlg_button_x_step(labels, width, &gap, &margin, &step))
-	++width;
-    width += (4 * MARGIN);
-    if (width > COLS)
-	width = COLS;
-    if (width > *limit)
-	*limit = width;
+    if (labels != 0 && dlg_button_count(labels)) {
+	while (!dlg_button_x_step(labels, width, &gap, &margin, &step))
+	    ++width;
+	width += (4 * MARGIN);
+	if (width > COLS)
+	    width = COLS;
+	if (width > *limit)
+	    *limit = width;
+    }
 }
 
 /*
@@ -257,7 +261,7 @@ dlg_draw_buttons(WINDOW *win,
     int final_y;
     int gap;
     int margin;
-    unsigned need;
+    size_t need;
     char *buffer;
 
     dlg_mouse_setbase(getbegx(win), getbegy(win));
@@ -277,7 +281,7 @@ dlg_draw_buttons(WINDOW *win,
     /*
      * Allocate a buffer big enough for any label.
      */
-    need = longest;
+    need = (size_t) longest;
     for (n = 0; labels[n] != 0; ++n) {
 	need += strlen(labels[n]) + 1;
     }
@@ -363,7 +367,7 @@ dlg_char_to_button(int ch, const char **labels)
     if (labels != 0) {
 	int j;
 
-	ch = dlg_toupper(dlg_last_getc());
+	ch = (int) dlg_toupper(dlg_last_getc());
 	for (j = 0; labels[j] != 0; ++j) {
 	    int cmp = dlg_button_to_char(labels[j]);
 	    if (ch == cmp) {
@@ -495,7 +499,8 @@ dlg_ok_labels(void)
     static const char *labels[5];
     int n = 0;
 
-    labels[n++] = my_ok_label();
+    if (!dialog_vars.nook)
+	labels[n++] = my_ok_label();
     if (dialog_vars.extra_button)
 	labels[n++] = my_extra_label();
     if (!dialog_vars.nocancel)
@@ -513,9 +518,9 @@ int
 dlg_ok_buttoncode(int button)
 {
     int result = DLG_EXIT_ERROR;
-    int n = 1;
+    int n = !dialog_vars.nook;
 
-    if (button <= 0) {
+    if (!dialog_vars.nook && (button <= 0)) {
 	result = DLG_EXIT_OK;
     } else if (dialog_vars.extra_button && (button == n++)) {
 	result = DLG_EXIT_EXTRA;
