@@ -1,35 +1,56 @@
 /*
- * $Id: inputstr.c,v 1.40 2004/12/23 00:46:15 tom Exp $
+ * $Id: inputstr.c,v 1.66 2010/01/15 23:13:36 tom Exp $
  *
- *  inputstr.c -- functions for input/display of a string
+ * inputstr.c -- functions for input/display of a string
  *
- * Copyright 2000-2003,2004	Thomas E. Dickey
+ * Copyright 2000-2009,2010 Thomas E. Dickey
  *
- *  This program is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU General Public License
- *  as published by the Free Software Foundation; either version 2
- *  of the License, or (at your option) any later version.
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU Lesser General Public License, version 2.1
+ *  as published by the Free Software Foundation.
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ *  This program is distributed in the hope that it will be useful, but
+ *  WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *  Lesser General Public License for more details.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  You should have received a copy of the GNU Lesser General Public
+ *  License along with this program; if not, write to
+ *	Free Software Foundation, Inc.
+ *	51 Franklin St., Fifth Floor
+ *	Boston, MA 02110, USA.
  */
 
-#include "dialog.h"
+#include <dialog.h>
+#include <dlg_keys.h>
+
 #include <errno.h>
 
 #ifdef HAVE_SETLOCALE
 #include <locale.h>
 #endif
 
+#if defined(HAVE_SEARCH_H) && defined(HAVE_TSEARCH)
+#include <search.h>
+#else
+#undef HAVE_TSEARCH
+#endif
+
+#ifdef NEED_WCHAR_H
+#include <wchar.h>
+#endif
+
+#if defined(USE_WIDE_CURSES)
+#define USE_CACHING 1
+#elif defined(HAVE_XDIALOG)
+#define USE_CACHING 1		/* editbox really needs caching! */
+#else
+#define USE_CACHING 0
+#endif
+
 typedef struct _cache {
     struct _cache *next;
-#ifdef USE_WIDE_CURSES
+#if USE_CACHING
     struct _cache *cache_at;	/* unique: associate caches by CACHE */
     const char *string_at;	/* unique: associate caches by char* */
 #endif
@@ -39,11 +60,15 @@ typedef struct _cache {
     int *list;			/* indices into the string */
 } CACHE;
 
-#ifdef USE_WIDE_CURSES
+#if USE_CACHING
 #define SAME_CACHE(c,s,l) (c->string != 0 && memcmp(c->string,s,l) == 0)
 
 static CACHE *cache_list;
 
+#ifdef HAVE_TSEARCH
+static void *sorted_cache;
+#endif
+
 #ifdef USE_WIDE_CURSES
 static int
 have_locale(void)
@@ -63,12 +88,55 @@ have_locale(void)
 }
 #endif
 
+#ifdef HAVE_TSEARCH
+static int
+compare_cache(const void *a, const void *b)
+{
+    const CACHE *p = (const CACHE *) a;
+    const CACHE *q = (const CACHE *) b;
+    int result = 0;
+    result = p->cache_at - q->cache_at;
+    if (result == 0)
+	result = p->string_at - q->string_at;
+    return result;
+}
+#endif
+
+static CACHE *
+find_cache(CACHE * cache, const char *string)
+{
+    CACHE *p;
+
+#ifdef HAVE_TSEARCH
+    void *pp;
+    CACHE find;
+
+    memset(&find, 0, sizeof(find));
+    find.cache_at = cache;
+    find.string_at = string;
+
+    if ((pp = tfind(&find, &sorted_cache, compare_cache)) != 0) {
+	p = *(CACHE **) pp;
+    } else {
+	p = 0;
+    }
+#else
+    for (p = cache_list; p != 0; p = p->next) {
+	if (p->cache_at == cache
+	    && p->string_at == string) {
+	    break;
+	}
+    }
+#endif
+    return p;
+}
+
 static void
 make_cache(CACHE * cache, const char *string)
 {
     CACHE *p;
 
-    p = (CACHE *) calloc(1, sizeof(CACHE));
+    p = dlg_calloc(CACHE, 1);
     assert_ptr(p, "load_cache");
     p->next = cache_list;
     cache_list = p;
@@ -77,6 +145,9 @@ make_cache(CACHE * cache, const char *string)
     p->string_at = string;
 
     *cache = *p;
+#ifdef HAVE_TSEARCH
+    (void) tsearch(p, &sorted_cache, compare_cache);
+#endif
 }
 
 static void
@@ -84,14 +155,11 @@ load_cache(CACHE * cache, const char *string)
 {
     CACHE *p;
 
-    for (p = cache_list; p != 0; p = p->next) {
-	if (p->cache_at == cache
-	    && p->string_at == string) {
-	    *cache = *p;
-	    return;
-	}
+    if ((p = find_cache(cache, string)) != 0) {
+	*cache = *p;
+    } else {
+	make_cache(cache, string);
     }
-    make_cache(cache, string);
 }
 
 static void
@@ -99,14 +167,10 @@ save_cache(CACHE * cache, const char *string)
 {
     CACHE *p;
 
-    for (p = cache_list; p != 0; p = p->next) {
-	if (p->cache_at == cache
-	    && p->string_at == string) {
-	    CACHE *q = p->next;
-	    *p = *cache;
-	    p->next = q;
-	    return;
-	}
+    if ((p = find_cache(cache, string)) != 0) {
+	CACHE *q = p->next;
+	*p = *cache;
+	p->next = q;
     }
 }
 #else
@@ -132,11 +196,11 @@ same_cache2(CACHE * cache, const char *string, unsigned i_len)
 	return TRUE;
     }
 
-    need = sizeof(int *) * (i_len + 1);
+    need = (i_len + 1);
     if (cache->list == 0) {
-	cache->list = malloc(need);
+	cache->list = dlg_malloc(int, need);
     } else if (cache->i_len < i_len) {
-	cache->list = realloc(cache->list, need);
+	cache->list = dlg_realloc(int, need, cache->list);
     }
     cache->i_len = i_len;
 
@@ -179,7 +243,7 @@ same_cache1(CACHE * cache, const char *string, unsigned i_len)
 
     return FALSE;
 }
-#endif /* USE_WIDE_CURSES */
+#endif /* USE_CACHING */
 
 /*
  * Counts the number of bytes that make up complete wide-characters, up to byte
@@ -198,27 +262,27 @@ dlg_count_wcbytes(const char *string, size_t len)
 	if (!same_cache1(&cache, string, len)) {
 	    while (len != 0) {
 		int part = 0;
-		int code = 0;
+		size_t code = 0;
 		const char *src = cache.string;
 		mbstate_t state;
-		int save = cache.string[len];
+		char save = cache.string[len];
 
 		cache.string[len] = '\0';
 		memset(&state, 0, sizeof(state));
 		code = mbsrtowcs((wchar_t *) 0, &src, len, &state);
 		cache.string[len] = save;
-		if (code >= 0) {
+		if ((int) code >= 0) {
 		    break;
 		}
 		++part;
 		--len;
 	    }
 	    cache.i_len = len;
+	    save_cache(&cache, string);
 	}
-	save_cache(&cache, string);
-	result = cache.i_len;
+	result = (int) cache.i_len;
     } else {
-	result = len;
+	result = (int) len;
     }
     return result;
 }
@@ -242,23 +306,23 @@ dlg_count_wchars(const char *string)
 	    const char *src = cache.string;
 	    mbstate_t state;
 	    int part = dlg_count_wcbytes(cache.string, len);
-	    int save = cache.string[part];
-	    int code;
-	    wchar_t *temp = calloc(sizeof(wchar_t), len + 1);
+	    char save = cache.string[part];
+	    size_t code;
+	    wchar_t *temp = dlg_calloc(wchar_t, len + 1);
 
 	    cache.string[part] = '\0';
 	    memset(&state, 0, sizeof(state));
-	    code = mbsrtowcs(temp, &src, part, &state);
-	    cache.i_len = (code >= 0) ? wcslen(temp) : 0;
+	    code = mbsrtowcs(temp, &src, (size_t) part, &state);
+	    cache.i_len = ((int) code >= 0) ? wcslen(temp) : 0;
 	    cache.string[part] = save;
 	    free(temp);
+	    save_cache(&cache, string);
 	}
-	save_cache(&cache, string);
-	result = cache.i_len;
+	result = (int) cache.i_len;
     } else
 #endif /* USE_WIDE_CURSES */
     {
-	result = strlen(string);
+	result = (int) strlen(string);
     }
     return result;
 }
@@ -271,7 +335,7 @@ const int *
 dlg_index_wchars(const char *string)
 {
     static CACHE cache;
-    unsigned len = dlg_count_wchars(string);
+    unsigned len = (unsigned) dlg_count_wchars(string);
     unsigned inx;
 
     load_cache(&cache, string);
@@ -285,7 +349,7 @@ dlg_index_wchars(const char *string)
 		mbstate_t state;
 		int width;
 		memset(&state, 0, sizeof(state));
-		width = mbrlen(current, strlen(current), &state);
+		width = (int) mbrlen(current, strlen(current), &state);
 		if (width <= 0)
 		    width = 1;	/* FIXME: what if we have a control-char? */
 		current += width;
@@ -294,11 +358,11 @@ dlg_index_wchars(const char *string)
 #endif /* USE_WIDE_CURSES */
 	    {
 		(void) current;
-		cache.list[inx] = inx;
+		cache.list[inx] = (int) inx;
 	    }
 	}
+	save_cache(&cache, string);
     }
-    save_cache(&cache, string);
     return cache.list;
 }
 
@@ -306,7 +370,7 @@ dlg_index_wchars(const char *string)
  * Given the character-offset to find in the list, return the corresponding
  * array index.
  */
-static int
+int
 dlg_find_index(const int *list, int limit, int to_find)
 {
     int result;
@@ -326,7 +390,7 @@ const int *
 dlg_index_columns(const char *string)
 {
     static CACHE cache;
-    unsigned len = dlg_count_wchars(string);
+    unsigned len = (unsigned) dlg_count_wchars(string);
     unsigned inx;
 
     load_cache(&cache, string);
@@ -339,41 +403,60 @@ dlg_index_columns(const char *string)
 	    mbstate_t state;
 
 	    for (inx = 0; inx < len; ++inx) {
-		wchar_t temp;
-		int check;
+		wchar_t temp[2];
+		size_t check;
 		int result;
 
-		memset(&state, 0, sizeof(state));
-		check = mbrtowc(&temp, string + inx_wchars[inx], num_bytes -
-				inx_wchars[inx], &state);
-		if (check <= 0)
-		    result = 1;
-		else
-		    result = wcwidth(temp);
-		if (result < 0) {
-		    cchar_t temp2;
-		    setcchar(&temp2, &temp, 0, 0, 0);
-		    result = wcslen(wunctrl(&temp2));
+		if (string[inx_wchars[inx]] == TAB) {
+		    result = ((cache.list[inx] | 7) + 1) - cache.list[inx];
+		} else {
+		    memset(&state, 0, sizeof(state));
+		    memset(temp, 0, sizeof(temp));
+		    check = mbrtowc(temp,
+				    string + inx_wchars[inx],
+				    num_bytes - (size_t) inx_wchars[inx],
+				    &state);
+		    if ((int) check <= 0) {
+			result = 1;
+		    } else {
+			result = wcwidth(temp[0]);
+		    }
+		    if (result < 0) {
+			wchar_t *printable;
+			cchar_t temp2;
+			setcchar(&temp2, temp, 0, 0, 0);
+			printable = wunctrl(&temp2);
+			result = printable ? (int) wcslen(printable) : 1;
+		    }
 		}
 		cache.list[inx + 1] = result;
-		if (inx > 0)
+		if (inx != 0)
 		    cache.list[inx + 1] += cache.list[inx];
 	    }
 	} else
 #endif /* USE_WIDE_CURSES */
 	{
 	    for (inx = 0; inx < len; ++inx) {
-		cache.list[inx + 1] = (isprint(UCH(string[inx]))
-				       ? 1
-				       : strlen(unctrl(UCH(string[inx]))));
-		if (string[inx] == '\n')
+		chtype ch = UCH(string[inx]);
+
+		if (ch == TAB)
+		    cache.list[inx + 1] =
+			((cache.list[inx] | 7) + 1) - cache.list[inx];
+		else if (isprint(ch))
 		    cache.list[inx + 1] = 1;
+		else {
+		    const char *printable;
+		    printable = unctrl(ch);
+		    cache.list[inx + 1] = (printable
+					   ? (int) strlen(printable)
+					   : 1);
+		}
 		if (inx != 0)
 		    cache.list[inx + 1] += cache.list[inx];
 	    }
 	}
+	save_cache(&cache, string);
     }
-    save_cache(&cache, string);
     return cache.list;
 }
 
@@ -390,7 +473,7 @@ dlg_count_columns(const char *string)
 	const int *cols = dlg_index_columns(string);
 	result = cols[limit];
     } else {
-	result = strlen(string);
+	result = (int) strlen(string);
     }
     return result;
 }
@@ -419,35 +502,19 @@ bool
 dlg_edit_string(char *string, int *chr_offset, int key, int fkey, bool force)
 {
     int i;
-    int len = strlen(string);
+    int len = (int) strlen(string);
     int limit = dlg_count_wchars(string);
     const int *indx = dlg_index_wchars(string);
     int offset = dlg_find_index(indx, limit, *chr_offset);
-    int max_len = MAX_LEN;
+    int max_len = dlg_max_input(MAX_LEN);
     bool edit = TRUE;
 
-    if (dialog_vars.max_input != 0 && dialog_vars.max_input < MAX_LEN)
-	max_len = dialog_vars.max_input;
-
     /* transform editing characters into equivalent function-keys */
     if (!fkey) {
 	fkey = TRUE;		/* assume we transform */
 	switch (key) {
 	case 0:
 	    break;
-	case CHR_BACKSPACE:
-	    key = KEY_BACKSPACE;
-	    break;
-	case 21:		/* ^U */
-	    key = KEY_DL;
-	    break;
-	case CHR_DELETE:
-	    key = KEY_DC;
-	    break;
-	case '\n':
-	case '\r':
-	    key = KEY_ENTER;
-	    break;
 	case ESC:
 	case TAB:
 	    fkey = FALSE;	/* this is used for navigation */
@@ -463,23 +530,23 @@ dlg_edit_string(char *string, int *chr_offset, int key, int fkey, bool force)
 	case 0:		/* special case for loop entry */
 	    edit = force;
 	    break;
-	case KEY_LEFT:
+	case DLGK_GRID_LEFT:
 	    if (*chr_offset)
 		*chr_offset = indx[offset - 1];
 	    break;
-	case KEY_RIGHT:
+	case DLGK_GRID_RIGHT:
 	    if (offset < limit)
 		*chr_offset = indx[offset + 1];
 	    break;
-	case KEY_HOME:
+	case DLGK_BEGIN:
 	    if (*chr_offset)
 		*chr_offset = 0;
 	    break;
-	case KEY_END:
+	case DLGK_FINAL:
 	    if (offset < limit)
 		*chr_offset = indx[limit];
 	    break;
-	case KEY_BACKSPACE:
+	case DLGK_DELETE_LEFT:
 	    if (offset) {
 		int gap = indx[offset] - indx[offset - 1];
 		*chr_offset = indx[offset - 1];
@@ -492,7 +559,7 @@ dlg_edit_string(char *string, int *chr_offset, int key, int fkey, bool force)
 		}
 	    }
 	    break;
-	case KEY_DC:
+	case DLGK_DELETE_RIGHT:
 	    if (limit) {
 		if (--limit == 0) {
 		    string[*chr_offset = 0] = '\0';
@@ -514,10 +581,24 @@ dlg_edit_string(char *string, int *chr_offset, int key, int fkey, bool force)
 		}
 	    }
 	    break;
-	case KEY_DL:
+	case DLGK_DELETE_ALL:
 	    string[*chr_offset = 0] = '\0';
 	    break;
-	case KEY_ENTER:
+	case DLGK_ENTER:
+	    edit = 0;
+	    break;
+#ifdef KEY_RESIZE
+	case KEY_RESIZE:
+	    edit = 0;
+	    break;
+#endif
+	case DLGK_GRID_UP:
+	case DLGK_GRID_DOWN:
+	case DLGK_FIELD_NEXT:
+	case DLGK_FIELD_PREV:
+	    edit = 0;
+	    break;
+	case ERR:
 	    edit = 0;
 	    break;
 	default:
@@ -525,13 +606,13 @@ dlg_edit_string(char *string, int *chr_offset, int key, int fkey, bool force)
 	    break;
 	}
     } else {
-	if (key == ESC || key == TAB) {
+	if (key == ESC || key == ERR) {
 	    edit = 0;
 	} else {
 	    if (len < max_len) {
 		for (i = ++len; i > *chr_offset; i--)
 		    string[i] = string[i - 1];
-		string[*chr_offset] = key;
+		string[*chr_offset] = (char) key;
 		*chr_offset += 1;
 	    } else {
 		(void) beep();
@@ -625,10 +706,16 @@ dlg_show_string(WINDOW *win,
 	    int check = cols[i + 1] - cols[scrollamt];
 	    if (check <= x_last) {
 		for (j = indx[i]; j < indx[i + 1]; ++j) {
-		    if (hidden && dialog_vars.insecure)
+		    chtype ch = UCH(string[j]);
+		    if (hidden && dialog_vars.insecure) {
 			waddch(win, '*');
-		    else
-			waddch(win, CharOf(string[j]));
+		    } else if (ch == TAB) {
+			int count = cols[i + 1] - cols[i];
+			while (--count >= 0)
+			    waddch(win, ' ');
+		    } else {
+			waddch(win, ch);
+		    }
 		}
 		k = check;
 	    } else {
@@ -646,8 +733,12 @@ dlg_show_string(WINDOW *win,
 void
 _dlg_inputstr_leaks(void)
 {
+#if USE_CACHING
     while (cache_list != 0) {
 	CACHE *next = cache_list->next;
+#ifdef HAVE_TSEARCH
+	tdelete(cache_list, &sorted_cache, compare_cache);
+#endif
 	if (cache_list->string != 0)
 	    free(cache_list->string);
 	if (cache_list->list != 0)
@@ -655,5 +746,6 @@ _dlg_inputstr_leaks(void)
 	free(cache_list);
 	cache_list = next;
     }
+#endif /* USE_CACHING */
 }
-#endif
+#endif /* NO_LEAKS */
