@@ -1,9 +1,9 @@
 /*
- *  $Id: ui_getc.c,v 1.40 2007/07/04 12:43:54 tom Exp $
+ *  $Id: ui_getc.c,v 1.48 2010/01/18 10:24:06 tom Exp $
  *
  * ui_getc.c - user interface glue for getc()
  *
- * Copyright 2001-2006,2007 Thomas E. Dickey
+ * Copyright 2001-2009,2010 Thomas E. Dickey
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License, version 2.1
@@ -71,6 +71,18 @@ dlg_add_callback(DIALOG_CALLBACK * p)
     wtimeout(p->win, WTIMEOUT_VAL);
 }
 
+/*
+ * Like dlg_add_callback(), but providing for cleanup of caller's associated
+ * state.
+ */
+void
+dlg_add_callback_ref(DIALOG_CALLBACK ** p, DIALOG_FREEBACK freeback)
+{
+    (*p)->caller = p;
+    (*p)->freeback = freeback;
+    dlg_add_callback(*p);
+}
+
 void
 dlg_remove_callback(DIALOG_CALLBACK * p)
 {
@@ -78,10 +90,13 @@ dlg_remove_callback(DIALOG_CALLBACK * p)
 
     if (p->input != 0) {
 	fclose(p->input);
+	if (p->input == dialog_state.pipe_input)
+	    dialog_state.pipe_input = 0;
 	p->input = 0;
     }
 
-    dlg_del_window(p->win);
+    if (!(p->keep_win))
+	dlg_del_window(p->win);
     if ((q = dialog_state.getc_callbacks) == p) {
 	dialog_state.getc_callbacks = p->next;
     } else {
@@ -93,6 +108,13 @@ dlg_remove_callback(DIALOG_CALLBACK * p)
 	    q = q->next;
 	}
     }
+
+    /* handle dlg_add_callback_ref cleanup */
+    if (p->freeback != 0)
+	p->freeback(p);
+    if (p->caller != 0)
+	*(p->caller) = 0;
+
     free(p);
 }
 
@@ -110,7 +132,9 @@ dlg_getc_ready(DIALOG_CALLBACK * p)
     FD_SET(fd, &read_fds);
 
     test.tv_sec = 0;		/* Seconds.  */
-    test.tv_usec = WTIMEOUT_VAL * 1000;		/* Microseconds.  */
+    test.tv_usec = (isatty(fd)	/* Microseconds.  */
+		    ? (WTIMEOUT_VAL * 1000)
+		    : 1);
     return (select(fd + 1, &read_fds, (fd_set *) 0, (fd_set *) 0, &test) == 1)
 	&& (FD_ISSET(fd, &read_fds));
 }
@@ -118,6 +142,7 @@ dlg_getc_ready(DIALOG_CALLBACK * p)
 int
 dlg_getc_callbacks(int ch, int fkey, int *result)
 {
+    int code = FALSE;
     DIALOG_CALLBACK *p, *q;
 
     if ((p = dialog_state.getc_callbacks) != 0) {
@@ -129,9 +154,9 @@ dlg_getc_callbacks(int ch, int fkey, int *result)
 		}
 	    }
 	} while ((p = q) != 0);
-	return TRUE;
+	code = (dialog_state.getc_callbacks != 0);
     }
-    return FALSE;
+    return code;
 }
 
 static void
@@ -236,7 +261,7 @@ dlg_getc(WINDOW *win, int *fkey)
 	    ch = ERR;
 	    *fkey = 0;
 	    code = wget_wch(win, &my_wint);
-	    my_wchar = my_wint;
+	    my_wchar = (wchar_t) my_wint;
 	    switch (code) {
 	    case KEY_CODE_YES:
 		ch = *fkey = my_wchar;
@@ -244,12 +269,12 @@ dlg_getc(WINDOW *win, int *fkey)
 		break;
 	    case OK:
 		memset(&state, 0, sizeof(state));
-		have_last_getc = wcrtomb(last_getc_bytes, my_wchar, &state);
+		have_last_getc = (int) wcrtomb(last_getc_bytes, my_wchar, &state);
 		if (have_last_getc < 0) {
 		    have_last_getc = used_last_getc = 0;
-		    last_getc_bytes[0] = my_wchar;
+		    last_getc_bytes[0] = (char) my_wchar;
 		}
-		ch = CharOf(last_getc_bytes[used_last_getc++]);
+		ch = (int) CharOf(last_getc_bytes[used_last_getc++]);
 		last_getc = my_wchar;
 		break;
 	    case ERR:
@@ -260,7 +285,7 @@ dlg_getc(WINDOW *win, int *fkey)
 		break;
 	    }
 	} else {
-	    ch = CharOf(last_getc_bytes[used_last_getc++]);
+	    ch = (int) CharOf(last_getc_bytes[used_last_getc++]);
 	}
 #else
 	ch = wgetch(win);
@@ -301,7 +326,7 @@ dlg_getc(WINDOW *win, int *fkey)
 		done = (interval <= 0);
 	    }
 	    if (!valid_file(stdin)
-		&& !valid_file(dialog_state.screen_output)) {
+		|| !valid_file(dialog_state.screen_output)) {
 		ch = ESC;
 		done = TRUE;
 	    }
@@ -331,6 +356,15 @@ dlg_getc(WINDOW *win, int *fkey)
 	    }
 	    /* FALLTHRU */
 	default:
+#ifdef NO_LEAKS
+	    if (before_lookup == DLG_CTRL('P')) {
+		/* for testing, ^P closes the connection */
+		close(0);
+		close(1);
+		close(2);
+		break;
+	    }
+#endif
 	    if ((p = dialog_state.getc_redirect) != 0) {
 		if (!(p->handle_getc(p, ch, *fkey, &result))) {
 		    dlg_remove_callback(p);
@@ -341,13 +375,6 @@ dlg_getc(WINDOW *win, int *fkey)
 		done = TRUE;
 	    }
 	    break;
-#ifdef NO_LEAKS
-	case DLG_CTRL('P'):	/* for testing, ^P closes the connection */
-	    close(0);
-	    close(1);
-	    close(2);
-	    break;
-#endif
 #ifdef HAVE_DLG_TRACE
 	case CHR_TRACE:
 	    dlg_trace_win(win);
